module Trying
  class Proxy < BasicObject
    attr_accessor :cell

    def initialize(obj)
      @cell = obj
    end

    def method_missing(method, *argv, &block)
      return nil unless @cell.respond_to?(method)

      @cell = @cell.instance_eval do
        public_send(method, *argv, &block)
      end
    end

    def with(&block)
      @cell = tap(&block)
    end

    def tap(&block)
      block.call(@cell)
    rescue => e
      @cell = e
    end

    def pipe(method, *argv, &block)
      method.call(@cell, *argv, &block)
    end

    def failed?(error = ::StandardError, &block)
      with(&block) if @cell.is_a? error
    end
  end

  def trying(&block)
    proxy = Proxy.new(self)
    bound = binding()
    result = nil

    obj = proxy.instance_eval do
      proxy = bound.local_variable_get(:proxy)
      proxy.cell = proxy.instance_eval(&block)
      bound.local_variable_set(:proxy, proxy)
    end

    obj.cell
  end
end
